#!/usr/bin/python
# -*- coding: ascii -*-
###########################################################################
# simple-gecrypt-decrypt.py - Simplified implementation of gecrypt decryption
#
# Copyright (C) 2007-2008 Dwayne C. Litzenberger <dlitz@dlitz.net>
# All rights reserved.
# 
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation.
# 
# THE AUTHOR PROVIDES THIS SOFTWARE ``AS IS'' AND ANY EXPRESSED OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Country of origin: Canada
#
###########################################################################
# Version 0.5 - 2008-08-04
###########################################################################

import sys
import os
import struct
import os.path
from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA256

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))   # PBKDF2.py is in the parent directory
from PBKDF2 import PBKDF2

# Definitions

def abort(error):
    print >>sys.stderr, "error: " + error
    sys.exit(1)

def read_bytes(bytes, file=sys.stdin):
    data = file.read(bytes)
    if len(data) != bytes:
        abort("premature end of file")
    return data

EXPECTED_FILE_ID = "616d1d67ca294e2eb98bc01ff0470300".decode('hex')

# Get input from the user
if sys.argv[1:]:
    keyfile = sys.argv[1]
else:
    print >>sys.stderr, "Usage: simple-gecrypt-decrypt.py keyfile < inputfile > outputfile"
    sys.exit(1)

# Read passphrase
passphrase = open(keyfile, "rb").read()

# Read FILE_ID
file_id = read_bytes(16)
if file_id != EXPECTED_FILE_ID:
    abort("unrecognized file format")

# Read the rest of the header
file_header = file_id + read_bytes(48)

# Parse nonce and iteration count
nonce = file_header[16:48]
iteration_count = struct.unpack("!H", file_header[48:50])[0]

# Compute the authentication and cipher keys
derived_octets = PBKDF2(passphrase, file_header, iteration_count, digestmodule=SHA256).read(112)
mac_key = derived_octets[:64]
cipher_key = derived_octets[64:96]
cipher_iv = derived_octets[96:112]

# Start hashing
h = HMAC.new(mac_key, digestmod=SHA256)
h.update(file_header)

# Start decryption
cipher = AES.new(cipher_key, AES.MODE_CBC, cipher_iv)

# Read chunks
while True:
    # Read a ciphertext block
    ciphertext_block = read_bytes(AES.block_size)   # read 16 bytes
    h.update(ciphertext_block)
    
    # Decrypt the block
    block = cipher.decrypt(ciphertext_block)

    # Decode the length field
    length_field = struct.unpack("!H", block[:2])[0]
    payload_size = length_field & 0x7fff
    ignore = bool(length_field & 0x8000)

    # Add the first plaintext block to the buffer
    PTXT = [block[2:]]
    bytes_remaining = payload_size - len(block[2:])
    while bytes_remaining > 0:
        # Read a ciphertext block
        ciphertext_block = read_bytes(AES.block_size)   # read 16 bytes
        h.update(ciphertext_block)
        
        # Decrypt the block
        block = cipher.decrypt(ciphertext_block)
        
        # Append the plaintext block to the buffer
        PTXT.append(block)

        # Subtract from bytes_remaining
        bytes_remaining -= len(block)

    PAYLOAD = "".join(PTXT)[:payload_size]
    
    # Read and check the MAC
    mac = read_bytes(h.digest_size) # read 32 bytes
    if mac != h.digest():
        abort("authentication failed")
    h.update(mac)

    # Output the payload, unless the ignore bit is set
    if not ignore:
        sys.stdout.write(PAYLOAD)
    
    # Check for EOF
    if not ignore and payload_size == 0:
        break
