#!/usr/bin/python
# -*- coding: ascii -*-
###########################################################################
# simple-gecrypt-encrypt.py - Simplified implementation of gecrypt encryption
#
# Copyright (C) 2007-2008 Dwayne C. Litzenberger <dlitz@dlitz.net>
# All rights reserved.
# 
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation.
# 
# THE AUTHOR PROVIDES THIS SOFTWARE ``AS IS'' AND ANY EXPRESSED OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Country of origin: Canada
#
###########################################################################
# Version 0.5 - 2008-08-04
###########################################################################

import sys
import os
import struct
from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA256

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))   # PBKDF2.py is in the parent directory
from PBKDF2 import PBKDF2

def make_chunk(payload, ignore):
    assert 1 <= len(payload) <= 32767
    if ignore:
        length_field = 0x8000 | len(payload)
    else:
        length_field = len(payload)
    retval = struct.pack("!H", length_field) + payload
    if len(retval) % 16 != 0:
        retval += "\0" * (16 - len(retval) % 16)
    return retval

FILE_ID = "616d1d67ca294e2eb98bc01ff0470300".decode('hex')

# Get input from the user
if sys.argv[1:]:
    keyfile = sys.argv[1]
else:
    print >>sys.stderr, "Usage: simple-gecrypt-encrypt.py keyfile [iteration-count] [pad-size] < inputfile > outputfile"
    sys.exit(1)

if sys.argv[2:]:
    iteration_count = int(sys.argv[2])
    if iteration_count < 1 or iteration_count > 0xffff:
        raise ValueError("iteration count must be between 1 and 65535")
else:
    iteration_count = 1000

if sys.argv[3:]:
    pad_size = int(sys.argv[3])
    if pad_size < 0:
        raise ValueError("padsize must be at least 0")
else:
    pad_size = 0

# Read master key
passphrase = open(keyfile, "rb").read()

# Randomly generate a 256-bit nonce
nonce = os.urandom(32)

# Generate and output the file header
file_header = FILE_ID + nonce + struct.pack("!H", iteration_count) + ("\0" * 14)
assert len(file_header) == 64

# Compute the authentication and cipher keys
derived_octets = PBKDF2(passphrase, file_header, iteration_count, digestmodule=SHA256).read(112)
mac_key = derived_octets[:64]
cipher_key = derived_octets[64:96]
cipher_iv = derived_octets[96:112]

# Start hashing
h = HMAC.new(mac_key, digestmod=SHA256)

# Output the file header
sys.stdout.write(file_header)
h.update(file_header)

# Start encryption
cipher = AES.new(cipher_key, AES.MODE_CBC, cipher_iv)

# Output random padding (this can be done anywhere, or even split up across the file)
pad_remaining = pad_size
while pad_remaining > 0:
    if pad_remaining > 32767:
        chunk_plaintext = make_chunk(os.urandom(32767), ignore=True)
        pad_remaining -= 32767
    else:
        chunk_plaintext = make_chunk(os.urandom(pad_remaining), ignore=True)
        pad_remaining = 0

    # Encrypt
    chunk_ciphertext = cipher.encrypt(chunk_plaintext)
    
    # Output the ciphertext (and update the hash state)
    sys.stdout.write(chunk_ciphertext)
    h.update(chunk_ciphertext)
    
    # Output the MAC (and update the hash state)
    mac = h.digest()
    sys.stdout.write(mac)
    h.update(mac)

# Output the encrypted input
payload = sys.stdin.read(32767)
while payload:
    chunk_plaintext = make_chunk(payload, ignore=False)
    
    # Encrypt
    chunk_ciphertext = cipher.encrypt(chunk_plaintext)
    
    # Output the ciphertext (and update the hash state)
    sys.stdout.write(chunk_ciphertext)
    h.update(chunk_ciphertext)
    
    # Output the MAC (and update the hash state)
    mac = h.digest()
    sys.stdout.write(mac)
    h.update(mac)

    # Read another block
    payload = sys.stdin.read(32767)

# Output the zero-length chunk (to indicate EOF)
chunk_plaintext = "\0" * 16

# Encrypt
chunk_ciphertext = cipher.encrypt(chunk_plaintext)

# Output the ciphertext (and update the hash state)
sys.stdout.write(chunk_ciphertext)
h.update(chunk_ciphertext)

# Output the MAC (and update the hash state)
mac = h.digest()
sys.stdout.write(mac)
h.update(mac)

# Done.
